	SUBROUTINE FILE_ACTION_ROUTINE
	IMPLICIT INTEGER (A-Z)
C++
C
C FUNCTION:
C
C	THIS ROUTINE IS CALLED AS AN ACTION ROUTINE BY FMG$FILE_SCAN
C	TO PROCESS EACH FILE IN THE WILD CARD CONTEXT
C
C INPUTS:
C
C	FAB		- FILE ACCESS BLOCK ( ALIAS FOR INPUT_FAB)
C	WORK_AREA	- CLI WORK AREA
C	OPEN_ERROR	- ADDRESS OF ROUTINE TO BE CALLED IF OPEN FAILS
C
C OUTPUTS:
C
C	THE FILE IS COPIED TO THE DESIGNATED LISTING FILE (EITHER IMPLICIT
C	OR EXPLICIT), LINE NUMBERS ARE ADDED IF /LINE_NUMBER QUALIFIER PRESENT,
C	AND AN ENTRY IS MADE IN INDEX IF /INDEX QUALIFIER DETECTED.
C
C--

	INCLUDE 'PARAM.INC'
	INCLUDE 'LENGTH.INC'
	INCLUDE 'RMSDATA.INC'
	INCLUDE 'STATUS.INC'
	INCLUDE 'BUFFER.INC'

	INTEGER*4 FILE_NAME_LENGTH
	INTEGER*4 NAME_DESCR(2)
	INTEGER*4 BUFFER_ADDRESS
	INTEGER*4 NAM_RSA
	INTEGER*4 RHB_ADDRESS

	INTEGER*2 LINE_NUMBER

	LOGICAL   PRINT_LINE_NUMBERS

	EXTERNAL OPEN_ERROR
	EXTERNAL MSC_RAB_ERROR
	EXTERNAL GET_ERROR
	EXTERNAL CLOSE_INPUT_ERROR

C
C	FORCE A PAGE SLEW ON NEXT WRITE TO OUTPUT FILE.
C

	CURRENT_LINE = MAX_LINES_PER_PAGE

C
C	OPEN INPUT FILE AND CONNECT RECORD ACCESS STREAM.
C

	IF(.NOT.SYS$OPEN(INPUT_FAB,OPEN_ERROR,)) THEN
	  RETURN
	  END IF

	IF(.NOT.SYS$CONNECT(INPUT_RAB,MSC_RAB_ERROR,)) THEN
	  CALL SYS$CLOSE(INPUT_FAB,CLOSE_INPUT_ERROR,)
	  RETURN
	  END IF

C
C	BUILD HEADING - FORMAT:
C
C		................COLUMN.................
C
C              0                               1       1
C              0                               2       3
C              1                               0       2
C FIELD
C
C HEADING_1	FILE NAME			PAGE	NNNNNN
C HEADING_2	(UNUSED)
C

C
C	EXTRACT LENGTH OF FILE NAME FROM NAM BLOCK
C

	CALL COPY(1,
	1         INPUT_NAM(NAM$B_RSL),
	2         FILE_NAME_LENGTH)

	CALL COPY(4,
	1         INPUT_NAM(NAM$L_RSA),
	2         NAM_RSA)

C
C	NOW FORMAT HEADING
C

	CALL SYS$FAO('!^!116<!AD!>    PAGE  ',	! CONTROL STRING
	1             LENGTH_HEADING_1,		! OUTLEN
	2             HEADING_1,		! OUTBUF
	3             %VAL(FILE_NAME_LENGTH),	! P1
	4             %VAL(NAM_RSA))

	CALL SYS$FAO('!/!/!/!/',		! CONTROL STRING
	1            LENGTH_HEADING_2,		! OUTLEN
	2            HEADING_2)			! OUTBUF

C
C	IF INDEX REQUESTED, ADD FILE INFO.
C

	IF((LIST_STATUS.AND.LIST_M_INDEX).NE.0) THEN

C
C	  BUILD DESCRIPTOR FOR FILE NAME
C

	  NAME_DESCR(1) = '010E0000'X .OR. FILE_NAME_LENGTH
	  NAME_DESCR(2) = NAM_RSA

C
C	  NOW ADD FILE NAME TO INDEX. ANY ERRORS WILL CAUSE INDEX
C	  BIT (LIST_V_INDEX) TO BE TURNED OFF.
C

	  CALL ADD_FILE_TO_INDEX(NAME_DESCR,
	1                        CURRENT_PAGE+1)	! +1 BECAUSE WE WILL DO SLEW FIRST
	  END IF

C
C	DETERMINE WHETHER OR NOT TO PRINT LINE NUMBERS. TO HAVE NUMBERS
C	PRINTED, FILE MUST BE VFC AND HAVE FSZ = 2, AND LINE NUMBER BIT
C	(LIST_V_LINE) MUST BE ON.
C

	IF(((LIST_STATUS .AND. LIST_M_LINE) .NE. 0) .AND. ! REQUESTED
	1  ((INPUT_FAB(FAB$B_RFM) .EQ. FAB$C_VFC) .AND.	! RIGHT RECORD FORMAT
	2   (INPUT_FAB(FAB$B_FSZ) .EQ. 2))) THEN	! RIGHT FIXED SIZE AREA

	  PRINT_LINE_NUMBERS = .TRUE.

	ELSE

	  PRINT_LINE_NUMBERS = .FALSE.
	  END IF

C
C	LOOP OVER ALL LINES IN INPUT FILE. COPY EACH LINE FROM
C	INPUT BUFFER TO OUTPUT BUFFER AND ADD LINE NUMBERS IF
C	NECESSARY.
C

	CALL COPY(4,
	1         INPUT_RAB(RAB$L_RHB),
	2         RAB_ADDRESS)

	DO WHILE (SYS$GET(INPUT_RAB,GET_ERROR,))

C
C	  EXTRACT LENGTH OF THIS RECORD FROM RAB.
C

	  CALL COPY(2,
	1           INPUT_RAB(RAB$W_RSZ),
	2           LENGTH_INPUT_RECORD)

C
C	  EXTRACT BUFFER ADDRESS (THIS MAY CHANGE SINCE WE ARE READING
C	  IN LOCATE MODE).
C

	  CALL COPY(4,
	1           INPUT_RAB(RAB$L_RBF),
	2           BUFFER_ADDRESS)

	  IF(PRINT_LINE_NUMBERS) THEN

C
C	    COPY LINE NUMBER FROM RHB.
C

	    CALL COPY(2,
	1             %VAL(RAB_ADDRESS),
	2             LINE_NUMBER)

	    CALL SYS$FAO('!5ZW   !AD',
	1                 LENGTH_OUTPUT_RECORD,
	2                 OUTPUT_BUFFER,
	3                 %VAL(LINE_NUMBER),
	4                 %VAL(LENGTH_INPUT_RECORD),
	5                 %VAL(BUFFER_ADDRESS))

	  ELSE

	    CALL SYS$FAO('        !AD',
	1                 LENGTH_OUTPUT_RECORD,
	2                 OUTPUT_BUFFER,
	3                 %VAL(LENGTH_INPUT_RECORD),
	4                 %VAL(BUFFER_ADDRESS))

	    END IF

	  CALL WRITE_LINE
	  END DO


	CALL SYS$CLOSE(INPUT_FAB,CLOSE_INPUT_ERROR,)
	RETURN
	END

