	INTEGER FUNCTION LIST_PRIN_OPT(CLI_REQUEST_BLOCK,
	1                              CLI_CALLBACK_ADDRESS,
	2                              STATUS)
	IMPLICIT INTEGER*4 (A-Z)
C++
C
C FUNCTION:
C
C	THIS MODULE IS CALLED AS A RESULT OF SEEING A /PRINTER
C	QUALIFIER.
C
C INPUTS:
C
C	CLI_REQUEST_BLOCK	CURRENT CLI REQUEST BLOCK
C	CLI_CALLBACK_ADDRESS	ADDRESS OF CLI UTILITY SERVICE ROUTINE
C	STATUS			STATUS (ALIAS FOR LIST_STATUS)
C
C IMPLICIT INPUTS:
C
C	OUTPUT_FAB		FAB FOR INPUT FILE
C
C OUTPUTS:
C
C	THE DEFAULT FILE NAME FOR THE LISTING FILE IS COPIED INTO
C	THE OUTPUT FAB. THIS FILE WILL SUBSEQUENTLY BE SPOOLED AND
C	DELETED.
C
C SIDE EFFECTS:
C
C	NONE
C
C--

	INCLUDE 'PARAM.INC'
	INCLUDE 'RMSDATA.INC'
	INCLUDE 'CLI.INC'

	INTEGER*4 FILE_OPTIONS
	INTEGER*4 DEVICE_SIZE
	INTEGER*4 DEVICE_DESCR(2)

	CHARACTER*18 SPOOL_FILE

	EXTERNAL LIST_BADQUENAM
	EXTERNAL LIST_LOGNAMERR

	EXTERNAL CREATE_ERROR
	EXTERNAL MSC_RAB_ERROR

	DATA SPOOL_FILE/'SYS$LOGIN:LIST.LIS'/

C
C	GET THE DEVICE NAME (CLI WILL SUPPLY DEFAULT VALUE OF SYS$PRINT
C	IF NO DEVICE IS EXPLICITLY SPECIFIED.
C

	LIST_PRIN_OPT = SYS$CLI(%VAL(REQUEST(GET_ASCII_VALUE)),
	1                       WORK_AREA,
	2                       CLI_REQUEST_BLOCK)

	IF(LIST_PRIN_OPT) THEN

C
C	  SUCCESSFUL PARSE OF DEVICE OR QUEUE NAME
C

	  CALL COPY(2,
	1           %VAL(REQUEST(GET_ASCII_VALUE)+CLI$W_RQSIZE),
	2           DEVICE_SIZE)

	  CALL COPY(4,
	1           %VAL(REQUEST(GET_ASCII_VALUE)+CLI$A_RQADDR),
	2           DEVICE_DESCR(2))

	  DEVICE_DESCR(1) = '010E0000'X .OR. DEVICE_SIZE

	  RETURN_STATUS = SYS$CRELOG(%VAL(2),
	1                            'SYS$PRINT',
	2                             DEVICE_DESCR,)

	  IF(.NOT.RETURN_STATUS) THEN

C
C	    WE COULDN'T CREATE THIS LOGICAL NAME
C

	    CALL LIB$SIGNAL(LIST_LOGNAMERR,
	1                   %VAL(2),
	2                   DEVICE_DESCR,
	3                   %VAL(RETURN_STATUS))
	    END IF

	ELSE

	  CALL LIB$SIGNAL(LIST_BADQUENAM,
	1                 %VAL(1),
	2                 %VAL(REQUEST(GET_ASCII_VALUE)+CLI$W_RQSIZE))

	  END IF

	CALL COPY(1,
	1         LEN(SPOOL_FILE),
	2         OUTPUT_FAB(FAB$B_FNS))

	CALL COPY(4,
	1         %LOC(SPOOL_FILE),
	2         OUTPUT_FAB(FAB$L_FNA))

C
C	SET THE FOP FIELD TO SPOOL AND DELETE THIS FILE.
C

	CALL COPY(4,
	1         OUTPUT_FAB(FAB$L_FOP),
	2         FILE_OPTIONS)

	FILE_OPTIONS = FILE_OPTIONS .OR. FAB$M_SPL .OR. FAB$M_DLT
	1              .OR. FAB$M_NAM

	CALL COPY(4,
	1         FILE_OPTIONS,
	2         OUTPUT_FAB(FAB$L_FOP))
C
C	  NOW TRY TO CREATE THIS FILE AND CONNECT A RECORD ACCESS
C	  STREAM.
C
	  CALL SYS$CREATE(OUTPUT_FAB,CREATE_ERROR,)
C
C	  A CREATE ERROR CAUSES AN IMMEDIATE UNWIND SO THE NEXT STATEMENT
C	  WOULD NOT BE EXECUTED IF THE CREATE FAILED.
C

	  CALL SYS$CONNECT(OUTPUT_RAB,MSC_RAB_ERROR,)

C
C	  THE MSC_RAB_ERROR IS NOT FATAL. THE USER WILL GET A WARNING;
C	  THE FIRST WRITE WILL CAUSE AN ABORT.
C

	RETURN
	END

	INTEGER FUNCTION LIST_INDEX_OPT(CLI_REQUEST_BLOCK,
	1                               CLI_CALLBACK_ADDRESS,
	2                               STATUS)
	IMPLICIT INTEGER (A-Z)
C++
C
C FUNCTION:
C
C	THIS ROUTINE IS CALLED WHEN A /INDEX QUALIFIER IS ENCOUNTERED.
C
C INPUTS:
C
C	CLI_REQUEST_BLOCK	CURRENT CLI REQUEST BLOCK
C	CLI_CALLBACK_ADDRESS	ADDRESS OF CLI UTILITY SERVICE ROUTINE
C	STATUS			STATUS (ALIAS FOR LIST_STATUS)
C
C IMPLICIT INPUTS:
C
C	NONE
C
C OUTPUTS:
C
C	SORT IS INITIALIZED
C
C SIDE EFFECTS:
C
C	IF AN ERROR IS ENCOUNTERED, THE CONDITION IS SIGNALED AND
C	THE INDEX BIT (LIST_V_INDEX) IS TURNED OFF.
C
C--

	INCLUDE 'STATUS.INC'
	INCLUDE 'LIBRARY:SS.DEF/NOLIST'

	INTEGER*4 RETURN_STATUS
	INTEGER*2 KEY_BUFFER(5)		! DEFINES SORT PARAMETERS

	EXTERNAL LIST_SORTERR

	DATA KEY_BUFFER/1,		! 1 KEY
	1               1,		! KEY TYPE = CHAR
	2               0,		! ASCENDING SEQUENCE
	3               1,		! STARTING KEY POSITION
	4              12/		! ENDING KEY POSITION

	RETURN_STATUS = SOR$INIT_SORT(KEY_BUFFER,
	1               144)		! LONGEST RECORD

	IF(.NOT. RETURN_STATUS) THEN

	  CALL LIB$SIGNAL(LIST_SORTERR,
	1                 %VAL(1),
	2                 %VAL(RETURN_STATUS))

	  END IF

	LIST_INDEX_OPT = SS$_NORMAL
	RETURN
	END

	INTEGER FUNCTION LIST_OUTP_OPT(CLI_REQUEST_BLOCK,
	1                              CLI_CALLBACK_ADDRESS,
	2                              STATUS)
	IMPLICIT INTEGER*4 (A-Z)
C++
C
C FUNCTION:
C
C	THIS MODULE PROCESSES A /OUTPUT QUALIFIER.
C
C INPUTS:
C
C	CLI_REQUEST_BLOCK	CURRENT CLI REQUEST BLOCK
C	CLI_CALLBACK_ADDRESS	ADDRESS OF CLI UTILITY SERVICE ROUTINE
C	STATUS			STATUS (ALIAS FOR LIST_STATUS)
C
C IMPLICIT INPUTS:
C
C	OUTPUT_FAB		FAB FOR OUTPUT FILE
C
C OUTPUTS:
C
C	THE FILE SPECIFICATION FOLLOWING /OUTPUT IS RETRIEVED
C	AND COPIED INTO OUTPUT FAB.
C
C SIDE EFFECTS:
C
C	ANY ERRORS ARE SIGNALED. THE RESULTING UNWIND CAUSES THE
C	ORIGINAL CALL BACK TO FAIL.
C
C--

	INCLUDE 'PARAM.INC'
	INCLUDE 'RMSDATA.INC'
	INCLUDE 'CLI.INC'

	EXTERNAL CREATE_ERROR
	EXTERNAL MSC_RAB_ERROR

C
C	DO ANOTHER CLI CALL BACK TO GET FILE NAME
C

	LIST_OUTP_OPT = SYS$CLI(%VAL(REQUEST(GET_ASCII_VALUE)),
	1                       WORK_AREA,
	2                       CLI_REQUEST_BLOCK)

	IF(LIST_OUTP_OPT) THEN

C
C	  WE WERE SUCCESSFUL IN GETING FILE NAME.
C

	  CALL COPY(1,
	1           %VAL(REQUEST(GET_ASCII_VALUE)+CLI$W_RQSIZE),
	2           OUTPUT_FAB(FAB$B_FNS))

	  CALL COPY(4,
	1           %VAL(REQUEST(GET_ASCII_VALUE)+CLI$A_RQADDR),
	2           OUTPUT_FAB(FAB$L_FNA))

C
C	  NOW TRY TO CREATE THIS FILE AND CONNECT A RECORD ACCESS
C	  STREAM.
C
	  CALL SYS$CREATE(OUTPUT_FAB,CREATE_ERROR,)
C
C	  A CREATE ERROR CAUSES AN IMMEDIATE UNWIND SO THE NEXT STATEMENT
C	  WOULD NOT BE EXECUTED IF THE CREATE FAILED.
C

	  CALL SYS$CONNECT(OUTPUT_RAB,MSC_RAB_ERROR,)

C
C	  THE MSC_RAB_ERROR IS NOT FATAL. THE USER WILL GET A WARNING;
C	  THE FIRST WRITE WILL CAUSE AN ABORT.
C

	  END IF

	RETURN
	END


	INTEGER FUNCTION LIST_DONE(CLI_REQUEST_BLOCK,
	1                    CLI_CALLBACK_ADDRESS,
	2                    STATUS)
	IMPLICIT INTEGER*4 (A-Z)
C++
C
C FUNCTION:
C
C	THIS ROUTINE IS CALLED WHEN THERE ARE NO MORE FILES
C	TO PROCESS. IF AN INDEX WAS REQUESTED, IT IS WRITTEN
C	TO THE OUTPUT FILE AT THIS TIME.
C
C INPUTS:
C
C	CLI_REQUEST_BLOCK	CURRENT CLI REQUEST BLOCK
C	CLI_CALLBACK_ADDRESS	ADDRESS OF CLI UTILITY SERVICE ROUTINE
C	STATUS			CLI STATUS AREA (ALIAS FOR LIST_STATUS)
C
C IMPLICIT INPUTS
C
C	OUTPUT_FAB		FILE ACCESS BLOCK FOR OUTPUT FILE
C	OUTPUT_RAB		RECORD ACCESS BLOCK FOR OUTPUT FILE
C
C OUTPUTS:
C
C	THE INDEX IS PRODUCED AND THE DONE BIT IS SET.
C
C SIDE EFFECTS:
C
C	A SORT ERROR CAUSES ANY REMAINING INDEX ENTRIES TO BE
C	DELETED. A WRITE ERROR CAUSES A COMMAND ABORT.
C
C--

	INCLUDE 'PARAM.INC'
	INCLUDE 'RMSDATA.INC'
	INCLUDE 'STATUS.INC'
	INCLUDE 'LENGTH.INC'
	INCLUDE 'BUFFER.INC'
	INCLUDE 'LIBRARY:SS.DEF/NOLIST'

	EXTERNAL LIST_SORTERR

C
C	SET DONE BIT.
C

	LIST_STATUS = LIST_STATUS .OR. LIST_M_DONE

C
C	SEE IF INDEX REQUESTED.
C

	IF((LIST_STATUS .AND. LIST_M_INDEX) .NE. 0) THEN

C
C	  INDEX WAS REQUESTED. SORT INDEX TABLE.
C

	  RETURN_STATUS=SOR$SORT_MERGE()

	  IF( RETURN_STATUS) THEN


C
C	    SUCCESSFUL SORT. CREATE INDEX HEADINGS.
C

	    CALL SYS$FAO('!^!117<INDEX OF STARTING PAGES!>PAGE I - ',
	1                LENGTH_HEADING_1,
	2                HEADING_1)

	    CALL SYS$FAO('!/!/!19<FILE NAME!>     !94<DEVICE:DIRECTORY!>' //
	1                'STARTING PAGE!/!/',
	2                LENGTH_HEADING_2,
	3                HEADING_2)

C
C	    FORCE PAGE SLEW AND RESET PAGE COUNT (INDEX BEGINS AT PAGE 1)
C

	    CURRENT_LINE = 55
	    CURRENT_PAGE = 0

C
C	    RETRIEVE SORTED RECORDS ONE AT A TIME.
C

	    DO WHILE (RETURN_STATUS)

	      RETURN_STATUS = SOR$RETURN_REC(OUTPUT_BUFFER,
	1                                    LENGTH_OUTPUT_RECORD)

	      IF(RETURN_STATUS) CALL WRITE_LINE
	      END DO

	    END IF

	    CALL SOR$END_SORT
	    IF(RETURN_STATUS .NE. SS$_ENDOFFILE) THEN

	      CALL LIB$SIGNAL(LIST_SORTERR,
	1                     %VAL(1),
	2                     %VAL(RETURN_STATUS))

	      END IF
	  END IF
	LIST_DONE = SS$_NORMAL
	RETURN
	END


