








                Vax Mp   -      Macro Processor



   Part I       1.      Introduction

   Part II      2.      What Mp does
                2.1         The pass
     The Mp     2.2         The macros
   Expander     2.3         Treatment of non macros
                2.4         Variables and Conditionals
                2.5         More internal statements
                2.6         Switches
                2.7         Macro definitions and arguments
                2.8         The string stack
                2.9         Macro argument operators
                3.      Operating instructions

   Part III     4.      Built in macros
		4.1         Language description
  Using the     4.2         Program structure.
Mp language     4.3         Data types
		4.4         Constants
		4.5         Declarative statements
		4.6         Executable statements
		4.7         Vax opcodes and Mp variables
                4.8         Notes
		4.9         Optimizer
                5.      Macro summary

   Part IV      6.      Alternate built in macro libraries
		6.1         Modifications to internals
                7.      Examples
.bp

.fo ''-#-'1.0'
.fo ''-#-'1.0'
1.  Introduction

Mp is a macro expander. This document describes the version implemented
on  the  pdp-11  (or  compatibility mode in the vax). It is intended to
support a systems  programming  language  for  the  Vax.  Since  the  C
language  is  not  universily available on Vax systems, and in at least
one version is not compatible with Vms external names; and Bliss  costs
money  and  is difficult to learn, programmers are still tempted to use
assembly language. Therefore, this simple macro processor  provides  an
alternative to macro-32 which can operate on data at the machine level,
while  still  supporting  higher  level language constructs. Part II of
this document deals with the use of Mp strictly as a  macro  processor;
Part  III  describes  an  implemented  language using Mp. The reader is
assumed to be familiar with macro-32 and especially the use of  macros.
The  documentation  of  the implemented language is short by design. In
any language which permits direct low  level  machine  operations,  the
user  must  be  able  to  predict  the  resultant  code each high level
statement generates, thus examination of the expanded code is a must.

The format of input  is  compatible  with  the  assembler  on  the  Vax
(Macro-32).  However,  the  arguments  inside  of macro definitions are
referenced by number rather than symbolically. Labels and comments  are
compatible  (i.e.  the  `:'  and  `;' delimiters for labels and comment
fields). One minor restriction is  that  the  `;'  must  have  a  space
preceding  it in order to begin the comment field. Parameters passed to
macros have the same nesting rules (i.e. <>'s etc.) but  there  are  no
Keyword  macros.  The  ^z...z syntax is supported and the form "..." is
additionally allowed. Arguments have selector  operators  which  permit
extracting  portions  of  a  single  argument  or  referencing multiple
arguments (e.g. arguments 2 thru arg  n,  where  n  is  the  number  of
arguments passed to the macro). 

In  addition,  there  is  a string stack which may be accessed globally
with operators for referencing absolute elements and elements  relative
to  the top of the stack. The stack may be divided into two stacks with
two stack pointers if  desired.  The  elements  on  the  stack  may  be
assigned  to  the  local parameters of the macro. A method is available
for generating labels of the form xxxnnn where nnn is a decimal  number
and  xxx is user definable (0-25 chars). There are 50 integer variables
30 of which can be accessed directly, and all 50 indirectly.  They  can
be  used  as  a numerical stack or as an index into the string stack. A
simplistic symbol table mechanisim exists  with  various  operators  to
access  the  type and/or existence of symbols. Various statements exist
for the conditional expansion of code either from within macros  or  at
the statement level.

The syntax of internal statements in this processor is cryptic, but for
a  good  reason  -  speed.  Any  interpreter that presents a high level
language must pay a price at translation time. The  language  which  is
implemented  using  this  system  can translate 250 lines of high level
code (on the vax with no other users) to intermediate form in about  25
seconds  (a  very  large module). And, since no macros are presented to
macro-32, it creates object modules in under 10  seconds.  (Why  5  rms
macros  take 25 seconds to expand is beyond the understanding of simple
mortals). 

The  language imlemented is a typeless statement language. It resembles
C with respect to storage class (auto, register, parameters, etc.)  and
supports  vax machine opcodes operating on scalar variables. Structures
and single dimension arrays are featured as well as a powerful function
call  (e.g.  func[a,b,func[...]])  which  can  include  expressions  or
character  string  literals.  Bit  arrays  are  supported using the vax
extract and  insert  bit  instructions.  For  convience,  an  intrinsic
function is supplied to create call by reference, whereas call by value
(using  stack  generated  arguments)  is  the norm. Except for cases of
deliberate abuse and the above call by reference routine (to  be  fixed
soon),  all  code  is  position  independent (unlike the Whitesmithes C
compiler). Scope is as in C, that is, there are local  stack  variables
in  separate  functions that optionally return values and external data
is sharable among all functions in a file but hidden from  the  outside
world.  Expressions are flat however, that is, no ()'s nor heirarchy of
operators. Direct manipulation  of  registers  are  permitted  and  the
translator   will   use   a  user  specified  range  of  registers  for
temporaries.  Since  the  language  is  essentially  just  a   macro-32
preprocessor,  macro32  features are all available. One problem using C
has been to reference external global constants  (e.g.  rms  structures
and  bit  definitions).  These usually require the creation of STRUCT's
which have to be carefully created and may need  recreation  at  a  new
release  of  VMS.  (Of  course  this is not C's fault, rather its DEC's
insistance on coding in assembler and charging big  bucks  for  bliss).
This  processor  can  be  used  for low level stuff (like the C runtime
system we wrote using C and MP for all the rms stuff) leaving the  high
level stuff to C. 

If the user does not directly violate the register usage rules, then an
optional optimizer will generate fairly good code. It is  not  globally
optimized,  but results in better code that Pascal and about as good or
better than the Whitesmithes C compiler. These measurements  are  based
soley  on  high  level coding, and don't attempt to measure what can be
accomplished using direct vax opcodes at critical  places.  The  output
code  is  not  commented,  but  uses  the  same symbolic names given to
declared variables. Thus if a user were forced to code at the  macro-32
level,  he  could use this translator and later comment the output code
and produce reasonable results. 

Other possible uses for this processor:  Partial  cross  assembly  from
macro11 to macro32 or cross assemblers for other machines.


.ne 4

.ne 10
.fo ''-#-'2.0'
2.  What Mp does



.ne 10
.fo ''-#-'2.1'
2.1  The pass

Mp  reads  the entire source file one time and acts as a filter copying
from its input to its output until it detects either an opcode that  is
an  internal  code or a user defined macro, or it detects an assignment
statement. Macros are recursively expanded (i.e. calls to the statement
routine for each line of the macro) until a line that is  not  a  macro
nor  an  internal opcode are detected, at which time the line is output
unmodified. Nesting  of  macro  calls  is  handled  much  the  same  as
macro-32. 

Note  that  mostly  all string matching has been made case insensitive;
where there are exceptions, they will be noted.



.ne 10
.fo ''-#-'2.2'
2.2  The macros

Mp  accepts  macros  from  various  places. The order in which they are
entered into the macro name table  determines  which  definition  takes
precedence  when the macro is multiply defined. The following lists the
the search order for opcodes. 


a.   The first set of names  are  the  conditionals  (these  are  hard-
     coded).

b.   The file which contains the source to be processed, may have macro
     definitions  (if  the  /lm  switch is specified - see later). They
     must precede all other source  code  and  be  separated  from  the
     source by a line with a slash `/' alone in column 1.

c.   One  macro  library may be specified and is a file of macros whose
     file-name is determined by Mp according the  following  rule.  The
     default  file-name is "[1,1]mplib.mac". The /ml switch changes the
     default name to "[your-directory]mplib.mac". A command-line option
     allows the setting of any file name in place of the defaults.  The
     /ml switch then is meaningless. 

d.   Built into Mp is a default set of macro definitions  (loaded  from
     an  overlay  makes processing faster). The loading of these may be
     suppressed by the /ni switch.

.ne 10
.fo ''-#-'2.3'
2.3  Treatment of non macros

Normally,  any statement not recognized by Mp is copied directly to the
output.  These  statements  may  be  any  Target  assembler  directives
provided  the  label: and ;comment syntax agrees. Note that the opcode
is what determines if the statement is a macro and that opcodes are not
limited to the rad50 subset. However, the current implementation of  Mp
has  additional  special case statement handling which is geared to the
language the author has devised using Mp. This special  checking  tests
for  assignment statements and simple machine opcodes such as clr, bit,
etc. This level of statement parsing only  occurs  at  statement  level
(i.e.  not  if  from  a  macro  expansion).  It  can be switched off if
necessary (see the /na switch later on).

.fo ''-#-'2.3.1'
2.3.1  Automatic macro calls

At  the  start  of  translation  of  a  file,  A  macro call is invoked
internally with the time and date as parameters. It is called as 

                $$TIM dd-mmm-yyy hh:mm:ss

The default library of macros will generate an  ident  statement  using
this  capability.  The user may redefine this to anything desirable. It
may be defined as  a  local  macro  which  will  override  the  default
definition. 



.ne 10
.fo ''-#-'2.4'
2.4 Variables and  Conditionals

     Mp  has  three  kinds  of  variables.  There  is  an  array of 50
     integers, a string array of 60 variable length strings, and  each
     macro  invocation  gets  20  strings.  There  are also 33 boolean
     switches. 


2.4.1 Integer variables

     There  is  an  array  of 50 integers the first 30 of which can be
     accessed as V0, V1, .. V9, va, vb, ...  vu  (vwxyz  reserved)  as
     well as by address (0..30). There are operations which permit the
     inclusion  of  these  variables  (see  section  on arguments) and
     others which set them (see s%n, t%yp, s%pos, and  s%len  internal
     statements).  They  may  also be accessed indirectly (as @n for n
     0..9,a..u) so they can form  a  stack.  Only  the  first  30  are
     directly  accessible so the remaining 20 must be accessed in this
     way also. The variable vz (readonly) allways contains the  number
     of arguments passed to a macro. 


2.4.2 Global string array variables

     There is a string array which can be manipulated  using  5  terse
     statements  (all begin with ?). This array of strings can be used
     as two stacks which is very useful for storing  nested  structure
     values such as for if statement generated labels. 

2.4.3 Local String variables

     Each  expansion  of  a macro gets 20 local string variables which
     are initialized to the actual parameters  of  the  called  macro.
     These operate recursively like high level language functions with
     local variables. Each invocation gets a new set. 

2.4.4 Switches (global)

     There are also 33 boolean switches. These may be set or reset and
     tested.  Many have fixed meanings, but may be set and tested just
     the same. They may be set from the command line but are initially
     all off. 


2.4.5 Conditionals

     A variety of conditional statements exist. There is a restriction
     on these; they cannot have labels on them. But, labels can appear
     on a line by themselves preceding these statements if desired. 

a.   I%B <string> <statement>  <else-statement>

         This expands statement if string is null (blank).

b.   I%NB <string> <statement>  <else-statement>

	 Expands statement if string is nonnull (nonblank).

c.   I%IDN <string1> <string2> <statement>  <else-statement>

         Expands statement if both strings are identical.
         (case is not significant). 

d.   I%DF  <string1> <string2> <statement>  <else-statement>

         Expands statement if strings are different. 
         (case is not significant). 

e.   I%GT <value1> <value2> <statement> <else-statement>

         Compares value1 to value2 and expands if greater than.
         The  values  must  be decimal numbers with an optional
         minus sign. Usually they reference some parameter that
         expands into a number.  This  is  also  the  way  that
         numerical variables are compared. 

	 For example:

		i%gt #@#a 50 <expand if a gt 50> <if not>

	 see below for explanation of #@#a.

f.   I%LT <value1> <value2> <statement> <else-statement>

	 Compares value1 to value2 and expands if less than.

g.   I%IN <key> <S1,S2,...,Sn> <E1,E2,...,En> <else>

         Compares key to each of Sn until  a  match  is  found,
         then  En  is  expanded. If no match is found, then the
         else statement is expanded. Else may be omitted.  Null
         strings  may  be present where desirable. (case is not
         significant). 


h.   I%SW switch# <statement1> <else-statement>

         Expands statement1 or the else statement depending  on
         whether  the  switch is set. See below for explanation
         of switches. 


i.   I%R <arg> <statement1> <else-statement>

         This statement tests arg for spaces, tabs, commas,  or
         square brackets and expands statement1 if any occur or
         the else-statement otherwise. Either may be null. This
         essentially  determines  if  the  arg is a `recursive'
         expression. 

	 Example:

		i%r <a[b]> <this, it is recursive> <not this>
		i%r <a-b.> <not this, it's not recursive> <but this>

j.   I%EXP <arg> <statement1> <else-statement>

         This  tests arg to see if it is an expression. It is a
         specific  test  for  the  default  macro  library.  An
         expression  is  anything  that  has multiple arguments
         where the second argument is one of the  4  arithmetic
         operators. 

         Example:


         I%EXP <a + b> <does this since there's a +> <not this>
         I%EXP <a[b]>  <not this> <but this, no expression>


k.  %if type parm1,parm2...
    %else
    %fi

         These three statements implement a block if conditional.
         The  %if  accepts  a  type  (from  the  single  line  if
         statements, less the i%) and any parameters depending  on
         the type. For example: 
.ne 10

            %if idn <#@1> <#@2>        %if in <#@1> <a,b,c,d>
               ...                        ...
            %else                      %fi
               ...
            %fi

              (The #@n notation is explained in section 2.7).

         These  may  be nested to any level, but must be the only
         conditional on a line. 

.ne 10
.fo ''-#-'2.5'
2.5 More internal statements

The following internal statements are also available.

a.   I%DOUT <statement>

         This will output statement  directly  without  further
         recursive  expansion.  This is useful when one desires
         to supplement code to  an  existing  target  assembler
         directive.  One  example  might be to output something
         additional to the .END statement by defining  a  macro
         called  ".END"  and having it output some code and end
         up with  an  output  of  the  actual  .END  directive.
         Comments may be output in this fashion as well. 

         Note that an equivalent form is:

                ~ statement

         where the `~' must be in column 1.

b.   I%ERR <message>

         This  outputs  a .error statement to the ti: device as
         well as to the file. The formats are:

             .error ;message           to the file
             line n message            to the terminal


c.   S%SW switch#

         This sets the switch specified. The switch  is  turned
         on  if  switch# is positive and off if preceded with a
         minus sign. 


d.   E%S <symbol> n {m}

         Enters symbol into a symbol  table  with  the  integer
         values of n and m (optionally) associated with it. The
         symbol may be subsequently looked up, see below. 

               Example:
                        e%s abc 5
			e%s def 5 4
         Places abc in symbol table with value  5. (The  symbol
         table  size  is a parameter that may be adjusted at mp
	 build time. See stak.mac).


e.   T%YP symbol number {number2}

         This directive places the type value (an integer) into
         the number register specified or zero if the symbol is
         not present. 

               Example:
                        t%yp abc 1
			t%yp def 1 2
         Places  the value of symbol abc (5 from above example)
         into variable v1. Def has two numbers  which  go  into
         variables v1 and v2. 

f.   D%SYM

         This  directive  dumps the symbol table as a series of
         macro calls. The format is as follows: 

                             $$$$$$4 n
                             $$$$$$5 sym val
                             $$$$$$6 n

         where  n  is  the  number of symbols and the $$$$$5 is
         expanded once for each symbol in the symbol table. The
         definitions for these are as debugging aids. They dump
         the  symbol  table   to   the   terminal   when   used
         interactively. 


g.   S%N num0 = num1 op num2

         Sets  the  value  of numeric variable num0 to num1 and
         optionally operates on it with num2.  The  valid  op's
         are: 

                        +  add
                        -  subtract
                        *  multiply
                        /  divide
                        &  logical and
                        !  logical or
                        L  lesser of the two
                        G  greater of the two

         Num1 and num2 are either signed  decimal  numbers  (16
         bits  worth) or v# to denote variables. There are only
         30 directly accessible numbers: 0..30. However,  using
         indirect  notation, 31-49 are accessible. This is done
         using @n notation (e.g. @9 means  indirect  thru  v9).
         Note that va's address is 10, vb's is 11, etc. 

         See also  the section on macro operands.

         Examples:

         S%N v0 = 6 + 7          ;sets numeric variable 0 to 13
	 S%N v0 = v6 * 7         ;sets v0 to v6's value * 7
         S%N v0 = 5              ;sets v0 to value 5
         S%N v6 = @5             ;sets v6 to value in variable
                                 ;pointed to by v5.
         S%N  @5 = 6             ;puts 6 into location v5 points
                                 ;to (indirect thru v5)
         s%n  v2 = 7 L 6         ;v2 gets the 6 (less than the 7)
	 s%n  vt = @a G vq       ;vt gets the greater of the number
				 ;pointed to by va and the contents
				 ;of vq
	 s%n  vt = vz - 1        ;vt gets the number of arguments less 1.

h.   S%len   <string>  num

         This computes the length  of  <string>  in  characters
         (may  be  zero)  and  puts  the  value  into numerical
         variable num. 
                E.g.
                        s%len <abc> 4    ;v4 gets a 3
			s%len <abc> 10   ;va gets a 3

i.   S%pos   <string>  <substring> num

         This  locates  the position of the substring in string
         and puts the value into numerical variable num. If the
         substring is not found, then zero  is  placed  in  the
         variable. 
		E.g.
			s%pos <abc> <ab> 4  ;v4 gets a 1
			s%pos <abc> <bc> 10 ;va gets a 2

         The  matching  is  case  insensitive.  If  any  of the
         substring character is a control-b (ascii 2)  it  acts
         as a single character wildcard matching anything. This
         character  is of course non printing and should not be
         used without good reason. 

j.   **  macro  parm1 parm2 ...

         This opcode expands `macro' once  for  each  non-blank
         parameters.  Note  that `macro' may be null to execute
         multiple commands on a single line. 

	 Example:

	      ** <>  s1,s2,s3    ;executes s1,s2,s3 in sequence
	      ** <abc def >  p1,p2,p3   ;executes:
						  abc def <p1>
						  abc def <p2>
						  abc def <p3>
					;in sequence

         Note  that `macro' can be any string; not only a macro
         name. Each parameter is concatonated  to  `macro'  and
         then expanded. When `macro' is null, a pair of <>'s is
	 NOT placed around Sn.

k.   %rept
     %endr
     %lexit

         The  most recent addition is a repeat block mechanism.
         It is not totally implemented however. The  %rept  and
         %endr  form  an  infinite loop (only within a macro!!)
         that must be broken via the %lexit (or a  %exit).  The
         problem  is that the %lexit does not take affect until
         the %endr is reached (if it exited  immediately,  then
         the  %if  level  would  get  lost). However, the %exit
         causes immediate exit of a macro (this  works  because
         macro  exit restores the %if level properly). In order
         to use properly, the %lexit should either  be  at  the
         end of the repeat block or used as follows: 

		%rept
		   ....                         ;stuff at top
		   %if ....                     ;test for exit
                      %lexit                    ;either exit
                   %else
                      ...                       ;or do rest of
                   %fi                          ;the loop
                %endr

l.   I%DIN  <statement>

         This  takes  statement  and  simulates  an  input from
         source  level.  This  means  it  will  check  for   an
         assignment   statement   etc.  The  reason  assignment
         statements are only checked at  source  level  is  for
         speed. 


m.  %exit

         This causes an immediate exit of the current macro.


n.  !mcr command

         This  is  primarily intended for interactive debugging
         of macro libraries.  It  permits  a  temporary  escape
         (like  in  the  editor)  to the operating system. NOTE
         that on  the  VAX  version  2.2  or  less  (at  least)
         spawning  from  a  compatibility mode process does not
         work and causes a fatal error such  that  the  process
         must be aborted.  The `!' must be in column 1.


.ne 10
.fo ''-#-'2.6'
2.6  Switches

     Mp reads the command line and extracts switches without  using
     the  CSI  modules.  Thus  switches  can appear anywhere on the
     line. They are always  2  letters  and  have  no  negation  or
     optional  data  capabilities.  Each switch is assigned a number
     (from 1 to 33) and it is this number which may  be  referenced
     in  the I%SW and S%SW statements described above. There are 16
     switches which are reserved for Mp and the rest can be used by
     the user. The switches are: 


1.   /nl

         No  .print statements generated. This is now the default
         since .print statements in macro-32 are  different  that
         in macro-11. 

2.   /da

	 Not implemented.

3.   /is

         This will print on ti: internal macros as they are read.
         This  aids  in  locating  a  faulty macro that bombs the
         preprocessor. 

4.   /me

         Expands  macros  one  level  deep  only  (for  debugging
         macros). 

5.   /ml

         Use local directory for standard macro  library  name,
         not [1,1].

6.   /op

         Not implemented.

7.   /tr

         Implements the trace feature. 

         This causes Mp to internally invoke two macro calls  for
         every  statement  input.  The two macro calls are passed
         the entire input line as  parameters.  One  call  occurs
         before expansion of the statement and one after. 

         Example:
                input line:   LET a = b

                affect:       $$$$$$1 LET a = b
                              (expansion of LET a = b)
                              $$$$$$2 LET a = b

         The two macros $$$$$$1 and $$$$$$2 must be user defined.

8.   /ni

         No internal macro definitions. 

9.   /md

         Expands   macros   and   outputs   (as  a  comment)  all
         intermediate steps during expansion of the macro. 

10.  /pp

         Not implemented. 

11.  /un

         Outputs to the ti: device, amounts of storage used after
         macro libraries are read  in  (to  determine  how  large
         table sizes should be in the task image - see stak.mac).

12.  /x2

         Turns  on unary operator recognition for first operator.
         It is not recommended that this switch be  applied  until
         all  macros  are  read  in,  or  any  detection  of  the
         operators during macro load time  will  cause  expansion
         which  is  not  usually desirable. Rather they should be
         turned on by program control as needed. See see  section
         on Macros for further information. 

13.  /x3

         Turns on unary operator recognition for second operator. 

14.  /x4

         Turns on unary operator recognition for third operator. 

15.  /x5

         Turns on unary operator recognition for quote operator.

31.  /lm

         Used  to  signal  that  local macros are included in the
         source file. 

32.  /na

         This directs the  statement  processor  to  not  attempt
         further  statement  parsing when the opcode field has an
         unknown symbol. 
33.  /nb

         This  suppresses the output of the stack usage unless it
         gets below a critically low value. 

34.  /mc

	 This  dumps macro call entrys and exits as comments.

.ne 30
         The  remaining  switches  are  unassigned  and  are   as
         follows: 


                switch 16 as /x6       <----|
                switch 17 as /x7            |
                switch 18 as /x8            |---{ used by default
                switch 19 as /x9            |     macros.
                switch 20 as /x0            |
                switch 21 as /y1       <----|
                switch 22 as /y2
                switch 23 as /y3
                switch 24 as /y4
                switch 25 as /y5
                switch 26 as /y6
                switch 27 as /y7
                switch 28 as /y8
                switch 29 as /y9       <----- used by default macros
                switch 30 as /y0



.fo ''-#-'2.7'
2.7  Macro definitions and arguments

     Macros have the form:

                       $MACRO name

                        macro text

                       $ENDM

Additional information on the $MACRO and $ENDM lines is optional and is
treated as comments.

Arguments  are referenced as #@n where n is 0..9,A..K (upper case). For
example, the following is a definition of a push macro. 

                        $MACRO push
                         mov #@1,-(sp)
                        $ENDM

This macro would be called (for example):

                        push this

And it would generate the following code:

                        mov this,-(sp)


One additional syntax item (not found in macro-32) is the following: An
argument of the form A[B] is treated as <A,B>.

Macro  arguments  are substituted into macro text as follows: A left to
right scan is performed copying text into a secondary buffer  character
for  character  until  an  argument reference occurs. Then the argument
text from the command line is copied replacing the argument  reference.
When  copying  the  argument text itself, no interpretation of argument
references is made and no  second  pass  on  the  line  is  made.  Thus
recursion within macro argument replacement is disallowed. 



.fo ''-#-'2.8'
2.8  The string stack


Mp has a string stack  which  is  communicated  with  using  5  special
directives  discussed below. These directives may be used anywhere in a
program either directly or inside a macro. However, when used outside a
macro there are no parameters to receive the values from this stack and
thus the results are discarded. 

The stack appears to be a 60 element string array.  The  first  24  are
generally  used  for  stack  1,  25-30 as a scratch area, and 31-60 for
stack 2. Item 29 contains the string used to generate labels. There are
2 forms of this construct as follows: 


  1.                        ??  p1,p2,...,pn

     This format does not  reference  the  string  stack  directly  but
     generates  automatic  symbols  of  the form $$$nnn and places them
     into the list of parameters specified. Example:

                               ?? 1,2

     would generate (assuming no other labels have yet been gen'd)  two
     symbols

                   $$$1 which would replace argument 1

                   $$$2 which would replace argument 2

     If the arguments  had  been  defined  (i.e.  not  null)  then  the
     previous  contents are lost.  The `$$$' is the default and by user
     control can be changed to some  other  string.  See  the  absolute
     string placement below.



.ne 6

  2.                 ?- <string> relative absolute switch

                                   and

                     ?+ argument relative absolute switch

     The  above  are  pushes  (-) and pops (+) of the string stack. The
     `?-' expects a literal string to be supplied as <string>. This  is
     usually done via some reference to an argument. 

     Example:

                               ?- #@1

     This will push argument 1 onto the stack.

     The second parameter (if supplied) specifies a  position  relative
     to  the  stack  pointer and ALSO denotes that the operation is NOT
     actually a push, rather it is a move.

     Stack relative placement does not modify the stack pointer.

     Example:

                              ?- #@1,2

     This will place argument 1 into the stack at position 2.  The  top
     grows downward and begins at location 1 of the string stack.

     The  third  parameter  (if  not  null)  specifies  that  the  push
     (actually move) is not relative to anything but is absolute.

     Example:

                             ?- abc,,29

     This will place the string `abc' into absolute position 29 of  the
     string   stack,   thereby  setting  the  label  generation  string
     described above.



     The fourth argument is to conditionalize the entire  operation  on
     the  status  of  a  particular  switch. This capability is not  as
     important as it was in previous versions of mp but it remains.

     Example:

                             ?- #@5,,,20

     This specifies that if switch 20 is set then push argument 5  onto
     the stack.

     To specify that the condition is based on the switch NOT being set
     then supply a negative switch number.

     The ?+ operation is analogous to the ?- operation with  respect  to
     the relative, absolute, and switch parameters. The first parameter
     is however the argument number in which to move to. This number is
     a decimal number.

     Example:

     Suppose we redefine the push  macro  such  that  if  the  argument
     supplied is reg0 then we will change it to R0.

     $MACRO push parm1
     ?- R0,,25                     ;put an R0 into string element 25
     I%IDN reg0,<#@1> <S%SW 26> <S%SW -26> ;set switch 26 if arg1 = `reg0'
     ?+ 1,,25,26                   ;if sw 26 set, move str 25 to arg1
     mov #@1,-(sp)                 ;now do the actual push
     $ENDM

     The above example demonstrates some of  the  stack  options  in  a
     macro  but  is  not  very  efficient  in  terms of statements (the
     overriding concern for speed). The following shows how  to  do  the
     same  thing  in one statement.


        $MACRO push
        I%IDN <#@1> reg0 <mov r0,-(sp)>  <mov #@1,-(sp)>
        $ENDM

     The second stack pointer is referenced as ++ or --.

     Example:

             ?-- abc         ;pushes "abc" onto stack 2.
             ?++ 3,2         ;pops next to top of stack 2 into parm 3

     Macros may also be expanded by the use of certain unary operators.
     When the scanner   encounters a token (each parameter of  a  macro
     and  the opcode itself) it checks to see if the first character is
     one of 3 operators (defined in the tkb file for mp). If it is  and
     switches  12,  13, or 14 respectively for operators 1, 2, and 3 is
     also set, then an expansion of the macro $x is performed (x is the
     unary operator). The token is stripped of the operator and  passed
     onto  the $x macro as parameters. The $x macro is user defined. It
     may generate code or just fiddle with the stack and switches  etc.
     On  exit  however, it MUST push onto stack one a string which will
     replace the token which caused the expansion. 
.ne 10

     Example:

                       LET a = _function[a,b]

     where  `_'  is  the  unary  operator  would  cause  the  following
     expansion to be invoked (before the LET is invoked): 

                       $_ function,a,b

     IF $_ returns (via a push on stack one) the  string  `replacement'
     then the let statement would be expanded as: 

                        LET a = replacement

     A suitable macro definition for `$_' could be: 

                        $MACRO $_
                        ?- replacement
                        $ENDM


.ne 20





     The  quote character (usually " but definable at mp build time) is
     a special case  unary  operator.  Its  interpretation  depends  on
     switch 15. If switch 15 is off, then a quoted string is treated as
     a macro argument in <>'s. For example: 

                        "Quoted string"

     is the same as:

                        <Quoted string>

     However,  if  switch  15 is on, then it is also treated as a unary
     operator but the entire argument is passed to the $" macro in  one
     piece. 

     For example:

                        LET a = "Quoted string"

     would cause "Quoted string" to be passed to $" as: 

                        $" <Quoted string>

     This  allows an implementation of the $" macro to output a labeled
     ascii statement (possibly in  a  data  psect)  and  replacing  the
     "Quoted  string"  (in  the  LET statement) with the address of the
     ascii string. This is basically what the default  macros  do.  See
     the file mcmd.mac. 

     Note that all the numbers specified in ?-, ?+, etc. statements can
     be macro parameters or numerical variables  as  discussed  in  the
     next section. 


.ne 10

.fo ''-#-'2.9'
2.9   Macro argument operators

 Macro arguments may be either simple argument references or  they  may
 select  portions  of the argument. An argument reference which has the
 following format selects a portion of the argument selected. 

                #@=n(f:l)

 In  this format `n' is the argument number, `f' is the first character
 in the argument and `l' is the last. The `n', `f', and `l'  parameters
 are  all  single digits 0..9,A..U or vx where x also can be 0..9,A..U.
 Anything not a number or letter (or `Z') (or larger than the  argument
 for the `l' parameter) is converted to the number of characters in the
 argument.  When the format is vx this means use the value contained in
 numerical variable x. 

 For example:


                #@=1(2:Z)       ;extracts chars 2-end of arg 1
                #@=A(B:D)       ;extracts chars 11-13 of arg 10
                #@=v3(v5:v6)    ;v3 specifies which argument, v5
                                ;specifies 1st character position
                                ;to extract, and v6 the last.

 To reference multiple arguments the following form is used:

                #@*(f:l)

 which  concatonates arguments `f' thru `l'. The same rules for f and l
 as above apply here as well except a value larger than the  number  of
 arguments  is  converted  to  the  number  of arguments (not including
 locally defined labels). 

 For example:

                #@*(2:Z)                ;selects args 2 thru n-args
                #@*(3:B)                ;selects args 3 thru 12
                #@*(v1:Z)               ;v1 thru last argument selected

 The `:' is actually a special case for the `separator' character which
 is output between each character  (in  extraction  of  characters)  or
 argument  (multiple argument selection). The following table describes
 the `separator' characters currently in use. 


.ne 7
  -------------------------------------------------------------------
  |     "       double quote            separates with ><            |
  |     :       colon                   no separation (null)         |
  |     _       underline character     separates with a space       |
  |     x       any other char          separates with x             |
  --------------------------------------------------------------------

.ne 10
For example, the macro call:    TEST arg1 arg2 arg3


The following argument references:



.ne 10
        #@=1(1,Z)       --->    a,r,g,1
        #@=1(1:Z)               arg1
        #@=1(1"Z)               a><r><g><1
        <#@=1(1"Z)>             <a><r><g><1>


        #@*(1,Z)                arg1,arg2,arg3
        #@*(1_Z)                arg1 arg2 arg3
        <#@*(1"2)>              <arg1><arg2>

 Note  that  arguments separated by <><> type pairs do not need a comma
 or space  between  them.  The  ><  characters  are  the  corresponding
 characters  defined in the Mp task build file for left and right angle
 brackets. If these values are  changed,  the  characters  used  as  ><
 separators in the argument selector also are changed. 

 A  special  case argument reference exists which permits conditionally
 appending (ap) or (fp) to  the  argument  (the  strings  appended  are
 defined in stak.mac). An argument reference of the form: 

                        #@{n}

 where n is of the form used above  to  reference  an  argument  number,
 causes  the  symbol  table  to  be  searched  for the string defined by
 argument n. Depending on  the  symbol  type  (auto  or  parameter)  the
 corresponding  register  is used in displacement mode. Additionally, if
 the symbol is of type register (10 thru 22), then it is  replaced  with
 the  corresponding  register Rx (computed via type-10). When the symbol
 table is searched, any leading `@'  is  ignored.  Thus,  @sym  will  be
 correctly  replaced  by @sym(ap), @sym(fp), or (Rx). This capability is
 heavily used by the default macros supplied with this processor. 

 This construct (#@{n}) also checks for an argument that  contains  only
 digits  (0..9) and an optional minus sign (-) which it then converts to
 #^D<x> to make it a decimal integer and to #^Ox if it is preceded by  a
 0.  Thus 9 becomes #^D<9> and 07 becomes #^O07. The <> are necessary to
 allow for negative numbers (#^D-1 must  be  #^D<-1>).  Lastly,  if  the
 argument  begins  with either ` or ' then it is mapped into #^a'...' or
 #^a`...`. The argument must be terminated by ` or ' as well. Thus `abc'
 is mapped into #^a`abc`.

 The 30 numerical variables discussed in the S%N internal statement can
 also be referenced inside a macro with a similar syntax. Any  argument
 of the form: 

                                #@#x     or
                                #@#@x

 Is  replaced  by  the  number  x  or  the number x points at (indirect
 through x). Recall that x may by 0..9,a..u.  Actually,  any  level  of
 indirection  may be specified with multiple @'s (e.g. #@#@@9 is double
 indirect). Be warned that  #@#vt  is  not  what  you  think  and  that
 variables v, w, x, y, z are reserved. 

                Example:

                        $MACRO test
			mov #@#t(fp),-(sp)
                        $ENDM


 In the above, it is assumed that vt contains an offset  into  register
 fp.  If  it  had  contained  the value 10 (decimal) then the following
 would have been output: 

                        mov 10(fp),-(sp)

 This is a useful mechanism for generating temporaries. 

.bp
.fo ''-#-'3.0'
3.  Operating instructions

     Mp uses the get-command-line routine with 1 level of  indirection.
     It does not make direct use of the CSI parsing routines; rather it
     parses the command line itself. The format is:

     mp>OUT=IN[+MACROLIB][/sw's]

             or

     mp>file[+macrolib][/sw's]

     IN is the input file (def ext=.mac)
     OUT is the output file (def ext = .mpp)
     MACROLIB is an optional library of MP macros (def ext = .mac)

     For example: To compile the file SOURCE.mac into SOURCE.obj:

		mp source           ;same as mp source.mpp=source.mac
		macro  source.mpp   ;macro-32 call


     If there is  no  `='  present,   then  "file"  is  converted  into
     "file=file".  In  this case the extensions must be defaulted.  The
     choice of .mpp for the intermediate file  extension  is  to  allow
     easy deletion of temporary files.

     If no command line is entered, repetitive mode is enabled. In this
     mode a file is translated until an eof on input. Then command mode
     is re-entered (like pip or tkb etc.). Thus  multiple  translations
     are permitted (usually from an indirect file). However, macros are
     only defined for the first command entered.  This allows a set  of
     modules to be translated with out the reloading of macro libraries
     (and the internal macro definitions). If new  macros  need  to  be
     defined the processor must be reloaded via a new invocation.

     One useful technique for debugging macros is the following:

         mcr>mp
             mp>ti:+lib/sw's          -input and output to tty
                 interactively test macros
                 control-z
             mp>ti:/sw's
                 more interactive tests with different switch settings
                 control-z
             mp>control-z             -to exit mp.

     This  technique allows one to test a macro and switch on the macro
     expansion if a problem results. It is also a good way to see how a
     macro expands before actually writing code that uses a  particular
     construct. One may also turn on/off switch 9 interactively to turn
     dumping on and off.

.ne 30

     Internal error reporting is poor in the expander. If messages come
     out out on  the  terminal  that  say  things  like  string  buffer
     overflow  or  string  index  out  of range, you have exceeded some
     internal limit. Check the last thing you wrote. Usually the  error
     is  caused by some butchered statement or one with unterminated <>
     pairs. The expander will write the offending line on the terminal.
     In some cases, a .error statement is generated. 


     Some times it may be necessary to rebuild the task with  a  larger
     stack and if many local symbols are generated there is a parameter
     which can be increased in the tkb file.  Stak.mac  also  has  some
     values which can be increased if you run out of space for macros.

.bp
.fo ''-#-'4.0'
